package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.citygml4j.core.model.core.AbstractCityObject;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

/**
 * Abstract base class for rooms inside CityGML 3.0 construction objects.
 */
public abstract class AbstractRoom extends CityObject {

    @Serial
    private static final long serialVersionUID = -1730625513988944329L;

    private final List<Installation> roomInstallations = new ArrayList<>(2);
    // Rooms have a Href list of furniture, the actual object is saved in the Building

    private final List<BoundarySurface> boundarySurfaceList = new ArrayList<>();


    protected org.citygml4j.core.model.core.AbstractUnoccupiedSpace cgmlRoom;


    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.accept(c);
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.accept(c);
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.collectContainedErrors(errors);
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.collectContainedErrors(errors);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.clearAllContainedCheckResults();
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.clearAllContainedCheckResults();
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (Installation roomInstallation : roomInstallations) {
            if (roomInstallation.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            if (boundarySurface.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (Installation roomInstallation : roomInstallations) {
            if (roomInstallation.containsAnyError()) {
                return true;
            }
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            if (boundarySurface.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            if (geom.getType() == GeometryType.MULTI_SURFACE) {
                MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                setMultiSurfaceAccordingToLod(geom, ms);
            } else {
                Solid solid = CityGmlUtils.createSolid(geom, factory, config);
                setSolidAccordingToLod(geom, solid);
            }
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.reCreateGeometries(factory, config);
        }
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.reCreateGeometries(factory, config);
        }
    }


    private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
        switch (geom.getLod()) {
            case LOD0:
                cgmlRoom.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                cgmlRoom.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                cgmlRoom.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to rooms");
        }
    }

    private void setSolidAccordingToLod(Geometry geom, Solid solid) {
        switch (geom.getLod()) {
            case LOD1:
                cgmlRoom.setLod1Solid(new SolidProperty(solid));
                break;
            case LOD2:
                cgmlRoom.setLod2Solid(new SolidProperty(solid));
                break;
            case LOD3:
                cgmlRoom.setLod3Solid(new SolidProperty(solid));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to rooms");
        }
    }


    @Override
    public void unsetGmlGeometries() {
        cgmlRoom.setLod0MultiSurface(null);
        cgmlRoom.setLod2MultiSurface(null);
        cgmlRoom.setLod3MultiSurface(null);
        cgmlRoom.setLod1Solid(null);
        cgmlRoom.setLod2Solid(null);
        cgmlRoom.setLod3Solid(null);
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.unsetGmlGeometries();
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.unsetGmlGeometries();
        }
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.prepareForChecking();
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (Installation roomInstallation : roomInstallations) {
            roomInstallation.clearMetaInformation();
        }
        for (BoundarySurface boundarySurface : boundarySurfaceList) {
            boundarySurface.clearMetaInformation();
        }
    }


    @Override
    public AbstractCityObject getGmlObject() {
        return cgmlRoom;
    }


    public List<Installation> getRoomInstallations() {
        return roomInstallations;
    }


    public List<BoundarySurface> getBoundarySurfaces() {
        return boundarySurfaceList;
    }

    public void addRoomInstallation(Installation roomInstallation) {
        roomInstallations.add(roomInstallation);
        roomInstallation.setParent(this);
    }

    public void addBoundarySurface(BoundarySurface boundarySurface) {
        boundarySurfaceList.add(boundarySurface);
        boundarySurface.setParent(this);
    }


    @Override
    public FeatureType getFeatureType() {
        return FeatureType.ROOM;
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        handler.addInstance(roomInstallations);
        handler.addInstance(boundarySurfaceList);
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        AbstractRoom originalAr = (AbstractRoom) original;
        for (BoundarySurface originalBs : originalAr.boundarySurfaceList) {
            boundarySurfaceList.add(handler.getCopyInstance(originalBs));
        }
        for (Installation originalRi : originalAr.roomInstallations) {
            roomInstallations.add(handler.getCopyInstance(originalRi));
        }
        cgmlRoom = originalAr.cgmlRoom;
    }

}
